/*
 * Copyright (c) 2016-2017 Satlab Aps <satlab@satlab.com>
 * All rights reserved. Do not distribute without permission.
 */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <inttypes.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <csp/csp_endian.h>

#include <slash/slash.h>

#include <satlab/polaris.h>
#include <prop-client/prop_client.h>

/* Default node and timeout */
static uint8_t polaris_node = SL_POLARIS_DEFAULT_ADDRESS;
static uint32_t polaris_timeout = 1000;

static uint8_t get_type(const uint8_t *data)
{
	return data[0] >> 2;
}

static uint32_t get_mmsi(const uint8_t *data)
{
	uint32_t mmsi;
	memcpy(&mmsi, &data[1], sizeof(mmsi));

	return csp_ntoh32(mmsi) >> 2;
}

static void print_hex(const uint8_t *data, size_t length)
{
	size_t i;

	for (i = 0; i < length; i++)
		printf("%02hhx", data[i]);
	printf("\n");
}

static int frameprint_cb(sl_polaris_store_frame_t *frame, void *arg)
{
	uint8_t type;
	uint32_t mmsi;

	type = get_type(frame->data);
	mmsi = get_mmsi(frame->data);

	printf("Frame %u ch: %u type: %2u mmsi: %9u scale: %u mean: %d\n",
		(unsigned)frame->seq,
		frame->channel,
		type,
		(unsigned)mmsi,
		frame->scale,
		frame->mean);
	print_hex(frame->data, frame->length);

	return 0;
}


/* Main commands */
slash_command_group(polaris, "Polaris subsystem commands");

static int cmd_polaris_node(struct slash *slash)
{
	uint8_t node;
	char *endptr;

	if (slash->argc < 2) {
		printf("Current node address: %hhu\n", polaris_node);
		return 0;
	}

	node = (uint8_t)strtoul(slash->argv[1], &endptr, 10);
	if (*endptr != '\0')
		return SLASH_EUSAGE;

	polaris_node = node;

	return 0;
}
slash_command_sub(polaris, node, cmd_polaris_node, "[address]",
		  "Set/read node address");

static int cmd_polaris_timeout(struct slash *slash)
{
	uint32_t timeout;
	char *endptr;

	if (slash->argc < 2) {
		printf("Current timeout: %"PRIu32" ms\n", polaris_timeout);
		return 0;
	}

	timeout = (uint32_t)strtoul(slash->argv[1], &endptr, 10);
	if (*endptr != '\0')
		return SLASH_EUSAGE;

	polaris_timeout = timeout;

	return 0;
}
slash_command_sub(polaris, timeout, cmd_polaris_timeout, "[ms]",
		  "Set/read communication timeout");


/* Bootloader commands */
slash_command_subgroup(polaris, boot, "Bootloader commands");

static int cmd_polaris_boot_flash(struct slash *slash)
{
	int ret;
	const char *filename;
	uint32_t crc;

	if (slash->argc < 2)
		return SLASH_EUSAGE;

	filename = slash->argv[1];

	printf("Flashing \"%s\" to alternate boot partition\n", filename);

	ret = sl_polaris_bootloader_flash_alternate(polaris_node, polaris_timeout, filename, &crc);
	if (ret < 0) {
		printf("Failed to flash image: %s\n", strerror(-ret));
		return SLASH_EIO;
	}

	printf("Flashed image, checksum = %08"PRIx32"\n", crc);

	return 0;
}
slash_command_subsub(polaris, boot, flash, cmd_polaris_boot_flash, "<filename>",
		     "Flash image from polaris SD-card to alternate partition");

static int cmd_polaris_boot_erase(struct slash *slash)
{
	int ret;

	ret = sl_polaris_bootloader_erase_alternate(polaris_node, polaris_timeout);
	if (ret < 0) {
		printf("Failed to erase image: %s\n", strerror(-ret));
		return SLASH_EIO;
	}

	printf("Success - alternate image was erased\n");

	return 0;
}
slash_command_subsub(polaris, boot, erase, cmd_polaris_boot_erase,
		     NULL, "Erase alternate partition");

static int cmd_polaris_boot_checksum(struct slash *slash)
{
	int ret;
	uint32_t size, crc;
	char *endptr;

	if (slash->argc < 2)
		return SLASH_EUSAGE;

	size = (uint32_t)strtoul(slash->argv[1], &endptr, 10);
	if (*endptr != '\0')
		return SLASH_EUSAGE;

	ret = sl_polaris_bootloader_checksum_alternate(polaris_node, polaris_timeout, size, &crc);
	if (ret < 0) {
		printf("Failed to checksum image: %s\n", strerror(-ret));
		return SLASH_EIO;
	}

	printf("Image checksum = %08"PRIx32"\n", crc);

	return 0;
}
slash_command_subsub(polaris, boot, checksum, cmd_polaris_boot_checksum,
		     "<size>", "Calculate checksum of alternate partition");

static int cmd_polaris_boot_set(struct slash *slash)
{
	int ret;
	uint8_t boots = 1;
	unsigned long boots_parsed;
	char *endptr;

	if (slash->argc > 1) {
		boots_parsed = strtoul(slash->argv[1], &endptr, 10);
		if (*endptr != '\0')
			return SLASH_EUSAGE;

		if (boots_parsed > 255) {
			fprintf(stderr, "Maximum number of boots is 255\n");
			return SLASH_EUSAGE;
		}

		boots = boots_parsed;
	}

	printf("Setting alternate firmware boot for %hhu boots\n", boots);

	ret = sl_polaris_bootloader_set_alternate(polaris_node, polaris_timeout, boots);
	if (ret < 0) {
		fprintf(stderr, "Failed to set boot image: %s\n", strerror(-ret));
		return SLASH_EIO;
	}

	printf("Success - Reboot to boot alternate image\n");

	return 0;
}
slash_command_subsub(polaris, boot, set, cmd_polaris_boot_set,
		     "[boots]", "Set number of boots on alternate partition");


/* Sample commands */
slash_command_subgroup(polaris, sample, "Sample buffer commands");

static int cmd_polaris_sample_read(struct slash *slash)
{
	int c, ret, fd, nbytes;
	size_t samples = 8192;
	uint8_t source = SL_POLARIS_SAMPLEBUF_SOURCE_RAW;
	uint32_t *sample_mem;

	while ((c = slash_getopt(slash, "s:n:")) != -1) {
		switch (c) {
		case 's':
			if (!strcmp(slash->optarg, "raw")) {
				source = SL_POLARIS_SAMPLEBUF_SOURCE_RAW;
			} else if (!strcmp(slash->optarg, "ch1")) {
				source = SL_POLARIS_SAMPLEBUF_SOURCE_CH1;
			} else if (!strcmp(slash->optarg, "ch2")) {
				source = SL_POLARIS_SAMPLEBUF_SOURCE_CH2;
			} else if (!strcmp(slash->optarg, "ch3")) {
				source = SL_POLARIS_SAMPLEBUF_SOURCE_CH3;
			} else if (!strcmp(slash->optarg, "ch4")) {
				source = SL_POLARIS_SAMPLEBUF_SOURCE_CH4;
			} else {
				source = atoi(slash->optarg);
			}
			break;
		case 'n':
			samples = atoi(slash->optarg);
			break;
		case '?':
		default:
			return SLASH_EUSAGE;
		}
	}

	if (slash->argc - slash->optind != 1)
		return SLASH_EUSAGE;

	fd = open(slash->argv[slash->optind], O_WRONLY | O_CREAT | O_TRUNC,
		  S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (fd < 0) {
		printf("Failed to open output file\n");
		return SLASH_EIO;
	}

	sample_mem = calloc(samples, sizeof(*sample_mem));
	if (!sample_mem) {
		printf("Failed to allocate memory\n");
		return SLASH_EIO;
	}

	ret = sl_polaris_sample_read(polaris_node, polaris_timeout, source, sample_mem, samples * sizeof(*sample_mem));
	if (ret < 0) {
		free(sample_mem);
		printf("Failed to read sample: %s\n", strerror(-ret));
		return SLASH_EIO;
	}

	nbytes = write(fd, sample_mem, samples * sizeof(*sample_mem));
	if (nbytes != samples * sizeof(*sample_mem)) {
		free(sample_mem);
		printf("Failed to write sample to output file\n");
	}

	free(sample_mem);

	return 0;
}
slash_command_subsub(polaris, sample, read, cmd_polaris_sample_read,
		     "[-s <source>] [-n <samples>] <filename>",
		     "Read sample buffer and write to file");

static int cmd_polaris_sample_store(struct slash *slash)
{
	int c, ret;
	size_t samples = 8192;
	uint8_t source = SL_POLARIS_SAMPLEBUF_SOURCE_RAW;
	uint8_t dest = SL_POLARIS_SAMPLEBUF_DEST_MEM;
	uint32_t id;
	char *endptr;

	while ((c = slash_getopt(slash, "s:d:n:")) != -1) {
		switch (c) {
		case 's':
			if (!strcmp(slash->optarg, "raw")) {
				source = SL_POLARIS_SAMPLEBUF_SOURCE_RAW;
			} else if (!strcmp(slash->optarg, "ch1")) {
				source = SL_POLARIS_SAMPLEBUF_SOURCE_CH1;
			} else if (!strcmp(slash->optarg, "ch2")) {
				source = SL_POLARIS_SAMPLEBUF_SOURCE_CH2;
			} else if (!strcmp(slash->optarg, "ch3")) {
				source = SL_POLARIS_SAMPLEBUF_SOURCE_CH3;
			} else if (!strcmp(slash->optarg, "ch4")) {
				source = SL_POLARIS_SAMPLEBUF_SOURCE_CH4;
			} else {
				source = atoi(slash->optarg);
			}
			break;
		case 'd':
			if (!strcmp(slash->optarg, "mem")) {
				dest = SL_POLARIS_SAMPLEBUF_DEST_MEM;
			} else if (!strcmp(slash->optarg, "sd")) {
				dest = SL_POLARIS_SAMPLEBUF_DEST_SDCARD;
			} else {
				dest = atoi(slash->optarg);
			}
			break;
		case 'n':
			samples = atoi(slash->optarg);
			break;
		case '?':
		default:
			return SLASH_EUSAGE;
		}
	}

	if (slash->argc - slash->optind != 1)
		return SLASH_EUSAGE;

	id = (uint32_t)strtoul(slash->argv[slash->optind], &endptr, 10);

	ret = sl_polaris_sample_store(polaris_node, polaris_timeout, source, dest, samples, id);
	if (ret < 0) {
		printf("Failed to store sample: %s\n", strerror(-ret));
		return SLASH_EIO;
	}

	return 0;
}
slash_command_subsub(polaris, sample, store, cmd_polaris_sample_store,
		     "[-s <source>] [-d <dest>] [-n <samples>] <id>",
		     "Save sample buffer on host");


/* Store commands */
slash_command_subgroup(polaris, store, "Store commands");

static int cmd_polaris_store_mirror(struct slash *slash)
{
	int ret;

	ret = sl_polaris_store_mirror_recv(polaris_node, polaris_timeout, frameprint_cb, NULL);
	if (ret < 0) {
		printf("Failed to request frames: %s\n", strerror(-ret));
		return SLASH_EIO;
	}

	return 0;
}
slash_command_subsub(polaris, store, mirror, cmd_polaris_store_mirror,
		     NULL, "Request mirror of received frames");

static int cmd_polaris_store_read(struct slash *slash)
{
	int c, ret, num = 0, bytes;
	uint8_t bitmap[SL_POLARIS_STORE_BITMAP_SIZE] = {0};
	uint32_t start = 0;
	char *endptr, *bitmapstr = NULL;
	size_t i, bitmapstrlen;

	while ((c = slash_getopt(slash, "b:n:s:")) != -1) {
		switch (c) {
		case 'b':
			bitmapstr = slash->optarg;
			break;
		case 's':
			start = (uint32_t)strtoul(slash->optarg, &endptr, 10);
			if (*endptr != '\0')
				return SLASH_EUSAGE;
			break;
		case 'n':
			num = (int)strtoul(slash->optarg, &endptr, 10);
			if (*endptr != '\0' || num < 1 || num > 8 * sizeof(bitmap))
				return SLASH_EUSAGE;
			break;
		case '?':
		default:
			return SLASH_EUSAGE;
		}
	}

	if (bitmapstr) {
		bitmapstrlen = strlen(bitmapstr);
		if (bitmapstrlen % 2 || bitmapstrlen / 2 > sizeof(bitmap))
			return SLASH_EUSAGE;

		bitmapstrlen = bitmapstrlen / 2;

		for (i = 0; i < bitmapstrlen; i++) {
			ret = sscanf(&bitmapstr[i * 2], "%2hhx", &bitmap[i]);
			if (ret != 1)
				return SLASH_EUSAGE;
		}
	}

	if (num > 0) {
		bytes = num / 8;
		memset(bitmap, 0xff, bytes);
		num -= bytes * 8;
		if (num > 0)
			bitmap[bytes] = ~((1 << (8 - num)) - 1);
	}

	ret = sl_polaris_store_read(polaris_node, polaris_timeout, start, bitmap, frameprint_cb, NULL);
	if (ret < 0) {
		printf("Failed to read frames: %s\n", strerror(-ret));
		return SLASH_EIO;
	}

	return 0;
}
slash_command_subsub(polaris, store, read, cmd_polaris_store_read,
		     "[-s start] [-n num] [-b bitmap]",
		     "Request frames from store");

static int cmd_polaris_run(struct slash *slash)
{
	int ret;
	const char *cmd;

	if (slash->argc != 2)
		return SLASH_EUSAGE;

	cmd = slash->argv[1];

	ret = sl_polaris_shell_run(polaris_node, polaris_timeout, cmd);
	if (ret < 0) {
		printf("Failed to run command: %s\n", strerror(-ret));
		return SLASH_EIO;
	}

	return 0;
}
slash_command_sub(polaris, run, cmd_polaris_run, "<command>",
		  "Run command on remote system");


/* Properties commands */
slash_command_subgroup(polaris, prop, "Properties commands");

static int cmd_polaris_prop_get(struct slash *slash)
{
	int ret;
	const struct prop_spec *spec;
	char valstr[128];

	if (slash->argc != 2)
		return SLASH_EUSAGE;

	spec = sl_prop_spec_find_by_name(&polaris_props, slash->argv[1]);
	if (!spec) {
		printf("No such property \"%s\"\n", slash->argv[1]);
		return SLASH_ENOENT;
	}

	ret = sl_prop_remote_get_to_string(polaris_node, polaris_timeout, spec, valstr, sizeof(valstr));
	if (ret < 0) {
		printf("Failed to read property value: %s\n", strerror(-ret));
		return SLASH_EINVAL;
	}

	printf("%s\n", valstr);

	return 0;
}
slash_command_subsub(polaris, prop, get, cmd_polaris_prop_get,
		     "<group.property>",
		     "Get remote properties");

static int cmd_polaris_prop_set(struct slash *slash)
{
	int ret;
	const struct prop_spec *spec;

	if (slash->argc != 3)
		return SLASH_EUSAGE;

	spec = sl_prop_spec_find_by_name(&polaris_props, slash->argv[1]);
	if (!spec) {
		printf("No such property \"%s\"\n", slash->argv[1]);
		return SLASH_ENOENT;
	}

	ret = sl_prop_remote_set_from_string(polaris_node, polaris_timeout, spec, slash->argv[2], strlen(slash->argv[2]));
	if (ret < 0) {
		if (ret == -EINVAL) {
			printf("Invalid value string\n");
		} else if (ret == -ENOENT) {
			printf("Invalid property index\n");
		} else if (ret == -ERANGE) {
			printf("Value is outside valid property range\n");
		} else if (ret == -ETIMEDOUT) {
			printf("Timeout while waiting for reply\n");
		} else {
			printf("Failed to set property value: %s\n", strerror(-ret));
		}
		return SLASH_EINVAL;
	}

	return 0;
}
slash_command_subsub(polaris, prop, set, cmd_polaris_prop_set,
		     "<group.property> <value>",
		     "Set remote property");

static int cmd_polaris_prop_reset(struct slash *slash)
{
	int ret;
	const struct prop_spec *spec;
	const struct prop_group_spec *group_spec;
	prop_id_t id;

	if (slash->argc < 2)
		return SLASH_EUSAGE;

	if (strchr(slash->argv[1], '.')) {
		spec = sl_prop_spec_find_by_name(&polaris_props, slash->argv[1]);
		if (!spec) {
			printf("No such property \"%s\"\n", slash->argv[1]);
			return SLASH_ENOENT;
		}

		id = spec->id;
	} else {
		group_spec = sl_prop_group_spec_find_by_name(&polaris_props, slash->argv[1]);
		if (!group_spec) {
			printf("No such property group \"%s\"\n", slash->argv[1]);
			return SLASH_ENOENT;
		}

		id = group_spec->id;
	}

	ret = sl_prop_remote_reset(polaris_node, polaris_timeout, id);
	if (ret < 0) {
		if (ret == -EINVAL) {
			printf("Invalid value string\n");
		} else if (ret == -ENOENT) {
			printf("Invalid property index\n");
		} else if (ret == -ERANGE) {
			printf("Value is outside valid property range\n");
		} else if (ret == -ETIMEDOUT) {
			printf("Timeout while waiting for reply\n");
		} else {
			printf("Failed to set property value: %s\n", strerror(-ret));
		}
		return SLASH_EINVAL;
	}

	return 0;
}
slash_command_subsub(polaris, prop, reset, cmd_polaris_prop_reset,
		     "<group>",
		     "Reset properties to default values");

static int polaris_prop_group_print_iter(const struct prop_group_spec *group, void *arg)
{
	printf("%-16s%-s\n", group->name, group->desc);

	return 0;
}

static int polaris_prop_print_iter(const struct prop_spec *prop, void *arg)
{
	char valstr[128] = "?";
	const char *type;
	const char *unit;
	struct sl_prop_query *query = (struct sl_prop_query *)arg;

	sl_prop_remote_query_get_reply_to_string(query, prop, valstr, sizeof(valstr));

	type = sl_prop_spec_type_string(prop);
	unit = prop->type.unit ? prop->type.unit : "";
	printf("%-16s%-8s%s %s\n", prop->name, type, valstr, unit);

	return 0;
}

static int cmd_polaris_prop_list(struct slash *slash)
{
	int ret;
	const struct prop_group_spec *group_spec;
	struct sl_prop_query query;
	
	uint8_t txbuf[16];
	uint8_t rxbuf[512];

	if (slash->argc < 2) {
		printf("\033[1m%-16s%-8s\033[0m\n", "Group", "Description");
		sl_prop_group_spec_iterate(&polaris_props, polaris_prop_group_print_iter, NULL);
	} else {
		group_spec = sl_prop_group_spec_find_by_name(&polaris_props, slash->argv[1]);
		if (!group_spec) {
			printf("No such property group \"%s\"\n", slash->argv[1]);
			return SLASH_ENOENT;
		}

		sl_prop_remote_query_create_static(&query, txbuf, sizeof(txbuf), rxbuf, sizeof(rxbuf));
		sl_prop_remote_query_get(&query, group_spec->id);
		ret = sl_prop_remote_query_send(&query, polaris_node, polaris_timeout);
		if (ret < 0) {
			fprintf(stderr, "Query failed: %s\n", strerror(-ret));
			return SLASH_EIO;
		}

		printf("\033[1m%-16s%-8s%s\033[0m\n", "Property", "Type", "Value");
		sl_prop_spec_iterate(group_spec, polaris_prop_print_iter, &query);
	}

	return 0;
}
slash_command_subsub(polaris, prop, list, cmd_polaris_prop_list,
		     "[group]",
		     "List property groups or group contents");

static int cmd_polaris_prop_save(struct slash *slash)
{
	int c, ret;
	const struct prop_group_spec *group;
	bool fallback = false;

	if (slash->argc < 2)
		return SLASH_EUSAGE;

	while ((c = slash_getopt(slash, "f")) != -1) {
		switch (c) {
		case 'f':
			fallback = true;
			break;
		default:
			return SLASH_EUSAGE;
		}
	}

	group = sl_prop_group_spec_find_by_name(&polaris_props, slash->argv[slash->optind]);
	if (!group) {
		printf("No such property group \"%s\"\n", slash->argv[slash->optind]);
		return SLASH_ENOENT;
	}

	ret = sl_prop_remote_save(polaris_node, polaris_timeout, group->id, fallback);
	if (ret < 0) {
		printf("Failed to save group: %s\n", strerror(-ret));
		return SLASH_EIO;
	}

	return 0;
}
slash_command_subsub(polaris, prop, save, cmd_polaris_prop_save,
		     "[-f] <group>", "Save property group");

static int cmd_polaris_prop_load(struct slash *slash)
{
	int c, ret;
	const struct prop_group_spec *group;
	bool fallback = false;

	if (slash->argc < 2)
		return SLASH_EUSAGE;

	while ((c = slash_getopt(slash, "f")) != -1) {
		switch (c) {
		case 'f':
			fallback = true;
			break;
		default:
			return SLASH_EUSAGE;
		}
	}

	group = sl_prop_group_spec_find_by_name(&polaris_props, slash->argv[slash->optind]);
	if (!group) {
		printf("No such property group \"%s\"\n", slash->argv[slash->optind]);
		return SLASH_ENOENT;
	}

	ret = sl_prop_remote_load(polaris_node, polaris_timeout, group->id, fallback);
	if (ret < 0) {
		printf("Failed to load group: %s\n", strerror(-ret));
		return SLASH_EIO;
	}

	return 0;
}
slash_command_subsub(polaris, prop, load, cmd_polaris_prop_load,
		     "[-f] <group>", "Load property group");

static int cmd_polaris_prop_erase(struct slash *slash)
{
	int c, ret;
	const struct prop_group_spec *group;
	bool fallback = false;

	if (slash->argc < 2)
		return SLASH_EUSAGE;

	while ((c = slash_getopt(slash, "f")) != -1) {
		switch (c) {
		case 'f':
			fallback = true;
			break;
		default:
			return SLASH_EUSAGE;
		}
	}

	group = sl_prop_group_spec_find_by_name(&polaris_props, slash->argv[slash->optind]);
	if (!group) {
		printf("No such property group \"%s\"\n", slash->argv[slash->optind]);
		return SLASH_ENOENT;
	}

	ret = sl_prop_remote_erase(polaris_node, polaris_timeout, group->id, fallback);
	if (ret < 0) {
		printf("Failed to erase group: %s\n", strerror(-ret));
		return SLASH_EIO;
	}

	return 0;
}
slash_command_subsub(polaris, prop, erase, cmd_polaris_prop_erase,
		     "[-f] <group>", "Erase property group");

static int cmd_polaris_prop_lock(struct slash *slash)
{
	int c, ret;
	const struct prop_group_spec *group;
	bool fallback = false;

	if (slash->argc < 2)
		return SLASH_EUSAGE;

	while ((c = slash_getopt(slash, "f")) != -1) {
		switch (c) {
		case 'f':
			fallback = true;
			break;
		default:
			return SLASH_EUSAGE;
		}
	}

	group = sl_prop_group_spec_find_by_name(&polaris_props, slash->argv[slash->optind]);
	if (!group) {
		printf("No such property group \"%s\"\n", slash->argv[slash->optind]);
		return SLASH_ENOENT;
	}

	ret = sl_prop_remote_lock(polaris_node, polaris_timeout, group->id, fallback);
	if (ret < 0) {
		printf("Failed to lock group: %s\n", strerror(-ret));
		return SLASH_EIO;
	}

	return 0;
}
slash_command_subsub(polaris, prop, lock, cmd_polaris_prop_lock,
		     "[-f] <group>", "Lock property group");

static int cmd_polaris_prop_unlock(struct slash *slash)
{
	int c, ret;
	const struct prop_group_spec *group;
	bool fallback = false;

	if (slash->argc < 2)
		return SLASH_EUSAGE;

	while ((c = slash_getopt(slash, "f")) != -1) {
		switch (c) {
		case 'f':
			fallback = true;
			break;
		default:
			return SLASH_EUSAGE;
		}
	}

	group = sl_prop_group_spec_find_by_name(&polaris_props, slash->argv[slash->optind]);
	if (!group) {
		printf("No such property group \"%s\"\n", slash->argv[slash->optind]);
		return SLASH_ENOENT;
	}

	ret = sl_prop_remote_unlock(polaris_node, polaris_timeout, group->id, fallback);
	if (ret < 0) {
		printf("Failed to unlock group: %s\n", strerror(-ret));
		return SLASH_EIO;
	}

	return 0;
}
slash_command_subsub(polaris, prop, unlock, cmd_polaris_prop_unlock,
		     "[-f] <group>", "Unlock property group");
